@@ -1,36 +1,24 @@
 /*
   This file is part of kcalcore library.
 
-  Copyright (c) 1998 Preston Brown <pbrown@kde.org>
-  Copyright (c) 2001 Cornelius Schumacher <schumacher@kde.org>
-  Copyright (c) 2002,2006 David Jarvie <djarvie@kde.org>
-  Copyright (C) 2005 Reinhold Kainhofer <kainhofer@kde.org>
+  SPDX-FileCopyrightText: 1998 Preston Brown <pbrown@kde.org>
+  SPDX-FileCopyrightText: 2001 Cornelius Schumacher <schumacher@kde.org>
+  SPDX-FileCopyrightText: 2002, 2006 David Jarvie <djarvie@kde.org>
+  SPDX-FileCopyrightText: 2005 Reinhold Kainhofer <kainhofer@kde.org>
 
-  This library is free software; you can redistribute it and/or
-  modify it under the terms of the GNU Library General Public
-  License as published by the Free Software Foundation; either
-  version 2 of the License, or (at your option) any later version.
-
-  This library is distributed in the hope that it will be useful,
-  but WITHOUT ANY WARRANTY; without even the implied warranty of
-  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-  Library General Public License for more details.
-
-  You should have received a copy of the GNU Library General Public License
-  along with this library; see the file COPYING.LIB.  If not, write to
-  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-  Boston, MA 02110-1301, USA.
+  SPDX-License-Identifier: LGPL-2.0-or-later
 */
 #include "recurrence.h"
-#include "utils_p.h"
 #include "recurrencehelper_p.h"
+#include "utils_p.h"
 
 #include "kcalendarcore_debug.h"
 
-#include <QDataStream>
-#include <QTimeZone>
 #include <QBitArray>
+#include <QDataStream>
 #include <QTime>
+#include <QTimeZone>
+#include <QHash>
 
 using namespace KCalendarCore;
 
@@ -39,21 +27,22 @@ class Q_DECL_HIDDEN KCalendarCore::Recurrence::Private
 {
 public:
     Private()
-        : mCachedType(rMax),
-          mAllDay(false),
-          mRecurReadOnly(false)
+        : mCachedType(rMax)
+        , mAllDay(false)
+        , mRecurReadOnly(false)
     {
     }
 
     Private(const Private &p)
-        : mRDateTimes(p.mRDateTimes),
-          mRDates(p.mRDates),
-          mExDateTimes(p.mExDateTimes),
-          mExDates(p.mExDates),
-          mStartDateTime(p.mStartDateTime),
-          mCachedType(p.mCachedType),
-          mAllDay(p.mAllDay),
-          mRecurReadOnly(p.mRecurReadOnly)
+        : mRDateTimes(p.mRDateTimes)
+        , mRDateTimePeriods(p.mRDateTimePeriods)
+        , mRDates(p.mRDates)
+        , mExDateTimes(p.mExDateTimes)
+        , mExDates(p.mExDates)
+        , mStartDateTime(p.mStartDateTime)
+        , mCachedType(p.mCachedType)
+        , mAllDay(p.mAllDay)
+        , mRecurReadOnly(p.mRecurReadOnly)
     {
     }
 
@@ -62,41 +51,37 @@ public:
     RecurrenceRule::List mExRules;
     RecurrenceRule::List mRRules;
     QList<QDateTime> mRDateTimes;
+    QHash<QDateTime, Period> mRDateTimePeriods; // Map RDate starts with periods if any
     DateList mRDates;
     QList<QDateTime> mExDateTimes;
     DateList mExDates;
-    QDateTime mStartDateTime;    // date/time of first recurrence
+    QDateTime mStartDateTime; // date/time of first recurrence
     QList<RecurrenceObserver *> mObservers;
 
     // Cache the type of the recurrence with the old system (e.g. MonthlyPos)
     mutable ushort mCachedType;
 
-    bool mAllDay = false;                // the recurrence has no time, just a date
+    bool mAllDay = false; // the recurrence has no time, just a date
     bool mRecurReadOnly = false;
 };
 
 bool Recurrence::Private::operator==(const Recurrence::Private &p) const
 {
-//   qCDebug(KCALCORE_LOG) << mStartDateTime << p.mStartDateTime;
-    if ((mStartDateTime != p.mStartDateTime &&
-            (mStartDateTime.isValid() || p.mStartDateTime.isValid())) ||
-            mAllDay != p.mAllDay ||
-            mRecurReadOnly != p.mRecurReadOnly ||
-            mExDates != p.mExDates ||
-            mExDateTimes != p.mExDateTimes ||
-            mRDates != p.mRDates ||
-            mRDateTimes != p.mRDateTimes) {
+    //   qCDebug(KCALCORE_LOG) << mStartDateTime << p.mStartDateTime;
+    if ((mStartDateTime != p.mStartDateTime && (mStartDateTime.isValid() || p.mStartDateTime.isValid())) || mAllDay != p.mAllDay
+        || mRecurReadOnly != p.mRecurReadOnly || mExDates != p.mExDates || mExDateTimes != p.mExDateTimes || mRDates != p.mRDates
+        || mRDateTimes != p.mRDateTimes || mRDateTimePeriods != p.mRDateTimePeriods) {
         return false;
     }
 
-// Compare the rrules, exrules! Assume they have the same order... This only
-// matters if we have more than one rule (which shouldn't be the default anyway)
+    // Compare the rrules, exrules! Assume they have the same order... This only
+    // matters if we have more than one rule (which shouldn't be the default anyway)
     int i;
     int end = mRRules.count();
     if (end != p.mRRules.count()) {
         return false;
     }
-    for (i = 0;  i < end;  ++i) {
+    for (i = 0; i < end; ++i) {
         if (*mRRules[i] != *p.mRRules[i]) {
             return false;
         }
@@ -105,7 +90,7 @@ bool Recurrence::Private::operator==(const Recurrence::Private &p) const
     if (end != p.mExRules.count()) {
         return false;
     }
-    for (i = 0;  i < end;  ++i) {
+    for (i = 0; i < end; ++i) {
         if (*mExRules[i] != *p.mExRules[i]) {
             return false;
         }
@@ -120,18 +105,19 @@ Recurrence::Recurrence()
 }
 
 Recurrence::Recurrence(const Recurrence &r)
-    : RecurrenceRule::RuleObserver(),
-      d(new KCalendarCore::Recurrence::Private(*r.d))
+    : RecurrenceRule::RuleObserver()
+    , d(new KCalendarCore::Recurrence::Private(*r.d))
 {
-    int i, end;
+    int i;
+    int end;
     d->mRRules.reserve(r.d->mRRules.count());
-    for (i = 0, end = r.d->mRRules.count();  i < end;  ++i) {
+    for (i = 0, end = r.d->mRRules.count(); i < end; ++i) {
         RecurrenceRule *rule = new RecurrenceRule(*r.d->mRRules[i]);
         d->mRRules.append(rule);
         rule->addObserver(this);
     }
     d->mExRules.reserve(r.d->mExRules.count());
-    for (i = 0, end = r.d->mExRules.count();  i < end;  ++i) {
+    for (i = 0, end = r.d->mExRules.count(); i < end; ++i) {
         RecurrenceRule *rule = new RecurrenceRule(*r.d->mExRules[i]);
         d->mExRules.append(rule);
         rule->addObserver(this);
@@ -160,7 +146,10 @@ Recurrence &Recurrence::operator=(const Recurrence &recurrence)
 
     // ### this copies the pointers in mExRules and mRRules eventually resulting in a double free!
     // fortunately however this function is unused, we just can't remove it just yet, due to ABI guarantees
+    QT_WARNING_PUSH
+    QT_WARNING_DISABLE_GCC("-Wdeprecated-copy")
     *d = *recurrence.d;
+    QT_WARNING_POP
     return *this;
 }
 #endif
@@ -194,10 +183,10 @@ void Recurrence::setAllDay(bool allDay)
     }
 
     d->mAllDay = allDay;
-    for (int i = 0, end = d->mRRules.count();  i < end;  ++i) {
+    for (int i = 0, end = d->mRRules.count(); i < end; ++i) {
         d->mRRules[i]->setAllDay(allDay);
     }
-    for (int i = 0, end = d->mExRules.count();  i < end;  ++i) {
+    for (int i = 0, end = d->mExRules.count(); i < end; ++i) {
         d->mExRules[i]->setAllDay(allDay);
     }
     updated();
@@ -227,7 +216,7 @@ void Recurrence::updated()
 {
     // recurrenceType() re-calculates the type if it's rMax
     d->mCachedType = rMax;
-    for (int i = 0, end = d->mObservers.count();  i < end;  ++i) {
+    for (int i = 0, end = d->mObservers.count(); i < end; ++i) {
         if (d->mObservers[i]) {
             d->mObservers[i]->recurrenceUpdated(this);
         }
@@ -255,9 +244,7 @@ ushort Recurrence::recurrenceType(const RecurrenceRule *rrule)
     RecurrenceRule::PeriodType type = rrule->recurrenceType();
 
     // BYSETPOS, BYWEEKNUMBER and BYSECOND were not supported in old versions
-    if (!rrule->bySetPos().isEmpty() ||
-            !rrule->bySeconds().isEmpty() ||
-            !rrule->byWeekNumbers().isEmpty()) {
+    if (!rrule->bySetPos().isEmpty() || !rrule->bySeconds().isEmpty() || !rrule->byWeekNumbers().isEmpty()) {
         return rOther;
     }
 
@@ -272,14 +259,11 @@ ushort Recurrence::recurrenceType(const RecurrenceRule *rrule)
     // BYMONTHDAY: with MONTHLY, YEARLY
     // BYMONTH: with YEARLY
     // BYYEARDAY: with YEARLY
-    if ((!rrule->byYearDays().isEmpty() && type != RecurrenceRule::rYearly) ||
-            (!rrule->byMonths().isEmpty() && type != RecurrenceRule::rYearly)) {
+    if ((!rrule->byYearDays().isEmpty() && type != RecurrenceRule::rYearly) || (!rrule->byMonths().isEmpty() && type != RecurrenceRule::rYearly)) {
         return rOther;
     }
     if (!rrule->byDays().isEmpty()) {
-        if (type != RecurrenceRule::rYearly &&
-                type != RecurrenceRule::rMonthly &&
-                type != RecurrenceRule::rWeekly) {
+        if (type != RecurrenceRule::rYearly && type != RecurrenceRule::rMonthly && type != RecurrenceRule::rWeekly) {
             return rOther;
         }
     }
@@ -344,11 +328,12 @@ bool Recurrence::recursOn(const QDate &qd, const QTimeZone &timeZone) const
         return false;
     }
 
-    int i, end;
+    int i;
+    int end;
     // For all-day events a matching exrule excludes the whole day
     // since exclusions take precedence over inclusions, we know it can't occur on that day.
     if (allDay()) {
-        for (i = 0, end = d->mExRules.count();  i < end;  ++i) {
+        for (i = 0, end = d->mExRules.count(); i < end; ++i) {
             if (d->mExRules[i]->recursOn(qd, timeZone)) {
                 return false;
             }
@@ -361,10 +346,10 @@ bool Recurrence::recursOn(const QDate &qd, const QTimeZone &timeZone) const
 
     // Check if it might recur today at all.
     bool recurs = (startDate() == qd);
-    for (i = 0, end = d->mRDateTimes.count();  i < end && !recurs;  ++i) {
+    for (i = 0, end = d->mRDateTimes.count(); i < end && !recurs; ++i) {
         recurs = (d->mRDateTimes[i].toTimeZone(timeZone).date() == qd);
     }
-    for (i = 0, end = d->mRRules.count();  i < end && !recurs;  ++i) {
+    for (i = 0, end = d->mRRules.count(); i < end && !recurs; ++i) {
         recurs = d->mRRules[i]->recursOn(qd, timeZone);
     }
     // If the event wouldn't recur at all, simply return false, don't check ex*
@@ -374,11 +359,11 @@ bool Recurrence::recursOn(const QDate &qd, const QTimeZone &timeZone) const
 
     // Check if there are any times for this day excluded, either by exdate or exrule:
     bool exon = false;
-    for (i = 0, end = d->mExDateTimes.count();  i < end && !exon;  ++i) {
+    for (i = 0, end = d->mExDateTimes.count(); i < end && !exon; ++i) {
         exon = (d->mExDateTimes[i].toTimeZone(timeZone).date() == qd);
     }
-    if (!allDay()) {       // we have already checked all-day times above
-        for (i = 0, end = d->mExRules.count();  i < end && !exon;  ++i) {
+    if (!allDay()) { // we have already checked all-day times above
+        for (i = 0, end = d->mExRules.count(); i < end && !exon; ++i) {
             exon = d->mExRules[i]->recursOn(qd, timeZone);
         }
     }
@@ -389,8 +374,8 @@ bool Recurrence::recursOn(const QDate &qd, const QTimeZone &timeZone) const
     } else {
         // Harder part: I don't think there is any way other than to calculate the
         // whole list of items for that day.
-//TODO: consider whether it would be more efficient to call
-//      Rule::recurTimesOn() instead of Rule::recursOn() from the start
+        // TODO: consider whether it would be more efficient to call
+        //      Rule::recurTimesOn() instead of Rule::recursOn() from the start
         TimeList timesForDay(recurTimesOn(qd, timeZone));
         return !timesForDay.isEmpty();
     }
@@ -402,23 +387,23 @@ bool Recurrence::recursAt(const QDateTime &dt) const
     const auto dtrecur = dt.toTimeZone(d->mStartDateTime.timeZone());
 
     // if it's excluded anyway, don't bother to check if it recurs at all.
-    if (std::binary_search(d->mExDateTimes.constBegin(), d->mExDateTimes.constEnd(), dtrecur) ||
-        std::binary_search(d->mExDates.constBegin(), d->mExDates.constEnd(), dtrecur.date())) {
+    if (std::binary_search(d->mExDateTimes.constBegin(), d->mExDateTimes.constEnd(), dtrecur)
+        || std::binary_search(d->mExDates.constBegin(), d->mExDates.constEnd(), dtrecur.date())) {
         return false;
     }
-    int i, end;
-    for (i = 0, end = d->mExRules.count();  i < end;  ++i) {
+    int i;
+    int end;
+    for (i = 0, end = d->mExRules.count(); i < end; ++i) {
         if (d->mExRules[i]->recursAt(dtrecur)) {
             return false;
         }
     }
 
     // Check explicit recurrences, then rrules.
-    if (startDateTime() == dtrecur ||
-        std::binary_search(d->mRDateTimes.constBegin(), d->mRDateTimes.constEnd(), dtrecur)) {
+    if (startDateTime() == dtrecur || std::binary_search(d->mRDateTimes.constBegin(), d->mRDateTimes.constEnd(), dtrecur)) {
         return true;
     }
-    for (i = 0, end = d->mRRules.count();  i < end;  ++i) {
+    for (i = 0, end = d->mRRules.count(); i < end; ++i) {
         if (d->mRRules[i]->recursAt(dtrecur)) {
             return true;
         }
@@ -440,7 +425,7 @@ QDateTime Recurrence::endDateTime() const
     if (!d->mRDateTimes.isEmpty()) {
         dts << d->mRDateTimes.last();
     }
-    for (int i = 0, end = d->mRRules.count();  i < end;  ++i) {
+    for (int i = 0, end = d->mRRules.count(); i < end; ++i) {
         auto rl = d->mRRules[i]->endDt();
         // if any of the rules is infinite, the whole recurrence is
         if (!rl.isValid()) {
@@ -540,20 +525,28 @@ void Recurrence::shiftTimes(const QTimeZone &oldTz, const QTimeZone &newTz)
     d->mStartDateTime = d->mStartDateTime.toTimeZone(oldTz);
     d->mStartDateTime.setTimeZone(newTz);
 
-    int i, end;
-    for (i = 0, end = d->mRDateTimes.count();  i < end;  ++i) {
-        d->mRDateTimes[i] = d->mRDateTimes[i].toTimeZone(oldTz);
-        d->mRDateTimes[i].setTimeZone(newTz);
+    QHash<QDateTime, Period> oldPeriods = d->mRDateTimePeriods;
+
+    for (auto &rDt : d->mRDateTimes) {
+        auto periodIt = oldPeriods.find(rDt);
+        periodIt->shiftTimes(oldTz, newTz);
+        rDt = rDt.toTimeZone(oldTz);
+        rDt.setTimeZone(newTz);
+        // Now there are QDateTime objects in the hash? is this shifting times?
+        d->mRDateTimePeriods.insert(rDt, *periodIt);
     }
-    for (i = 0, end = d->mExDateTimes.count();  i < end;  ++i) {
-        d->mExDateTimes[i] = d->mExDateTimes[i].toTimeZone(oldTz);
-        d->mExDateTimes[i].setTimeZone(newTz);
+
+    for (auto &exDt : d->mExDateTimes) {
+        exDt = exDt.toTimeZone(oldTz);
+        exDt.setTimeZone(newTz);
     }
-    for (i = 0, end = d->mRRules.count();  i < end;  ++i) {
-        d->mRRules[i]->shiftTimes(oldTz, newTz);
+
+    for (auto &rr : d->mRRules) {
+        rr->shiftTimes(oldTz, newTz);
     }
-    for (i = 0, end = d->mExRules.count();  i < end;  ++i) {
-        d->mExRules[i]->shiftTimes(oldTz, newTz);
+
+    for (auto exR : d->mExRules) {
+        exR->shiftTimes(oldTz, newTz);
     }
 }
 
@@ -578,6 +571,7 @@ void Recurrence::clear()
     d->mExRules.clear();
     d->mRDates.clear();
     d->mRDateTimes.clear();
+    d->mRDateTimePeriods.clear();
     d->mExDates.clear();
     d->mExDateTimes.clear();
     d->mCachedType = rMax;
@@ -605,13 +599,14 @@ void Recurrence::setStartDateTime(const QDateTime &start, bool isAllDay)
         return;
     }
     d->mStartDateTime = start;
-    setAllDay(isAllDay);               // set all RRULEs and EXRULEs
+    setAllDay(isAllDay); // set all RRULEs and EXRULEs
 
-    int i, end;
-    for (i = 0, end = d->mRRules.count();  i < end;  ++i) {
+    int i;
+    int end;
+    for (i = 0, end = d->mRRules.count(); i < end; ++i) {
         d->mRRules[i]->setStartDt(start);
     }
-    for (i = 0, end = d->mExRules.count();  i < end;  ++i) {
+    for (i = 0, end = d->mExRules.count(); i < end; ++i) {
         d->mExRules[i]->setStartDt(start);
     }
     updated();
@@ -840,7 +835,7 @@ void Recurrence::setMonthlyPos(const QList<RecurrenceRule::WDayPos> &monthlyDays
         return;
     }
 
-    //TODO: sort lists
+    // TODO: sort lists
     // the position inside the list has no meaning, so sort the list before testing if it changed
 
     if (monthlyDays != rrule->byDays()) {
@@ -867,7 +862,7 @@ void Recurrence::addMonthlyDate(short day)
     }
 }
 
-void Recurrence::setMonthlyDate(const QList< int > &monthlyDays)
+void Recurrence::setMonthlyDate(const QList<int> &monthlyDays)
 {
     if (d->mRecurReadOnly) {
         return;
@@ -900,7 +895,7 @@ void Recurrence::setYearly(int freq)
 // Daynumber within year
 void Recurrence::addYearlyDay(int day)
 {
-    RecurrenceRule *rrule = defaultRRule(false);   // It must already exist!
+    RecurrenceRule *rrule = defaultRRule(false); // It must already exist!
     if (!rrule) {
         return;
     }
@@ -914,7 +909,7 @@ void Recurrence::addYearlyDay(int day)
 
 void Recurrence::setYearlyDay(const QList<int> &days)
 {
-    RecurrenceRule *rrule = defaultRRule(false);   // It must already exist!
+    RecurrenceRule *rrule = defaultRRule(false); // It must already exist!
     if (!rrule) {
         return;
     }
@@ -997,8 +992,9 @@ void Recurrence::setYearlyMonth(const QList<int> &months)
 
 TimeList Recurrence::recurTimesOn(const QDate &date, const QTimeZone &timeZone) const
 {
-// qCDebug(KCALCORE_LOG) << "recurTimesOn(" << date << ")";
-    int i, end;
+    // qCDebug(KCALCORE_LOG) << "recurTimesOn(" << date << ")";
+    int i;
+    int end;
     TimeList times;
 
     // The whole day is excepted
@@ -1009,7 +1005,7 @@ TimeList Recurrence::recurTimesOn(const QDate &date, const QTimeZone &timeZone)
     // EXRULE takes precedence over RDATE entries, so for all-day events,
     // a matching excule also excludes the whole day automatically
     if (allDay()) {
-        for (i = 0, end = d->mExRules.count();  i < end;  ++i) {
+        for (i = 0, end = d->mExRules.count(); i < end; ++i) {
             if (d->mExRules[i]->recursOn(date, timeZone)) {
                 return times;
             }
@@ -1022,7 +1018,7 @@ TimeList Recurrence::recurTimesOn(const QDate &date, const QTimeZone &timeZone)
     }
 
     bool foundDate = false;
-    for (i = 0, end = d->mRDateTimes.count();  i < end;  ++i) {
+    for (i = 0, end = d->mRDateTimes.count(); i < end; ++i) {
         dt = d->mRDateTimes[i].toTimeZone(timeZone);
         if (dt.date() == date) {
             times << dt.time();
@@ -1031,14 +1027,14 @@ TimeList Recurrence::recurTimesOn(const QDate &date, const QTimeZone &timeZone)
             break; // <= Assume that the rdatetime list is sorted
         }
     }
-    for (i = 0, end = d->mRRules.count();  i < end;  ++i) {
+    for (i = 0, end = d->mRRules.count(); i < end; ++i) {
         times += d->mRRules[i]->recurTimesOn(date, timeZone);
     }
     sortAndRemoveDuplicates(times);
 
     foundDate = false;
     TimeList extimes;
-    for (i = 0, end = d->mExDateTimes.count();  i < end;  ++i) {
+    for (i = 0, end = d->mExDateTimes.count(); i < end; ++i) {
         dt = d->mExDateTimes[i].toTimeZone(timeZone);
         if (dt.date() == date) {
             extimes << dt.time();
@@ -1047,8 +1043,8 @@ TimeList Recurrence::recurTimesOn(const QDate &date, const QTimeZone &timeZone)
             break;
         }
     }
-    if (!allDay()) {       // we have already checked all-day times above
-        for (i = 0, end = d->mExRules.count();  i < end;  ++i) {
+    if (!allDay()) { // we have already checked all-day times above
+        for (i = 0, end = d->mExRules.count(); i < end; ++i) {
             extimes += d->mExRules[i]->recurTimesOn(date, timeZone);
         }
     }
@@ -1059,14 +1055,15 @@ TimeList Recurrence::recurTimesOn(const QDate &date, const QTimeZone &timeZone)
 
 QList<QDateTime> Recurrence::timesInInterval(const QDateTime &start, const QDateTime &end) const
 {
-    int i, count;
+    int i;
+    int count;
     QList<QDateTime> times;
-    for (i = 0, count = d->mRRules.count();  i < count;  ++i) {
+    for (i = 0, count = d->mRRules.count(); i < count; ++i) {
         times += d->mRRules[i]->timesInInterval(start, end);
     }
 
     // add rdatetimes that fit in the interval
-    for (i = 0, count = d->mRDateTimes.count();  i < count;  ++i) {
+    for (i = 0, count = d->mRDateTimes.count(); i < count; ++i) {
         if (d->mRDateTimes[i] >= start && d->mRDateTimes[i] <= end) {
             times += d->mRDateTimes[i];
         }
@@ -1074,7 +1071,7 @@ QList<QDateTime> Recurrence::timesInInterval(const QDateTime &start, const QDate
 
     // add rdates that fit in the interval
     QDateTime kdt = d->mStartDateTime;
-    for (i = 0, count = d->mRDates.count();  i < count;  ++i) {
+    for (i = 0, count = d->mRDates.count(); i < count; ++i) {
         kdt.setDate(d->mRDates[i]);
         if (kdt >= start && kdt <= end) {
             times += kdt;
@@ -1086,10 +1083,7 @@ QList<QDateTime> Recurrence::timesInInterval(const QDateTime &start, const QDate
     // mStartDateTime.
     // So, If we have rdates/rdatetimes but don't have any rrule we must explicitly
     // add mStartDateTime to the list, otherwise we won't see the first occurrence.
-    if ((!d->mRDates.isEmpty() || !d->mRDateTimes.isEmpty()) &&
-            d->mRRules.isEmpty() &&
-            start <= d->mStartDateTime &&
-            end >= d->mStartDateTime) {
+    if ((!d->mRDates.isEmpty() || !d->mRDateTimes.isEmpty()) && d->mRRules.isEmpty() && start <= d->mStartDateTime && end >= d->mStartDateTime) {
         times += d->mStartDateTime;
     }
 
@@ -1098,7 +1092,7 @@ QList<QDateTime> Recurrence::timesInInterval(const QDateTime &start, const QDate
     // Remove excluded times
     int idt = 0;
     int enddt = times.count();
-    for (i = 0, count = d->mExDates.count();  i < count && idt < enddt;  ++i) {
+    for (i = 0, count = d->mExDates.count(); i < count && idt < enddt; ++i) {
         while (idt < enddt && times[idt].date() < d->mExDates[i]) {
             ++idt;
         }
@@ -1108,7 +1102,7 @@ QList<QDateTime> Recurrence::timesInInterval(const QDateTime &start, const QDate
         }
     }
     QList<QDateTime> extimes;
-    for (i = 0, count = d->mExRules.count();  i < count;  ++i) {
+    for (i = 0, count = d->mExRules.count(); i < count; ++i) {
         extimes += d->mExRules[i]->timesInInterval(start, end);
     }
     extimes += d->mExDateTimes;
@@ -1123,8 +1117,8 @@ QDateTime Recurrence::getNextDateTime(const QDateTime &preDateTime) const
     // prevent infinite loops, e.g. when an exrule extinguishes an rrule (e.g.
     // the exrule is identical to the rrule). If an occurrence is found, break
     // out of the loop by returning that QDateTime
-// TODO_Recurrence: Is a loop counter of 1000 really okay? I mean for secondly
-// recurrence, an exdate might exclude more than 1000 intervals!
+    // TODO_Recurrence: Is a loop counter of 1000 really okay? I mean for secondly
+    // recurrence, an exdate might exclude more than 1000 intervals!
     int loop = 0;
     while (loop < 1000) {
         // Outline of the algo:
@@ -1175,8 +1169,8 @@ QDateTime Recurrence::getNextDateTime(const QDateTime &preDateTime) const
         nextDT = dates.first();
 
         // Check if that date/time is excluded explicitly or by an exrule:
-        if (!std::binary_search(d->mExDates.constBegin(), d->mExDates.constEnd(), nextDT.date()) &&
-            !std::binary_search(d->mExDateTimes.constBegin(), d->mExDateTimes.constEnd(), nextDT)) {
+        if (!std::binary_search(d->mExDates.constBegin(), d->mExDates.constEnd(), nextDT.date())
+            && !std::binary_search(d->mExDateTimes.constBegin(), d->mExDateTimes.constEnd(), nextDT)) {
             bool allowed = true;
             for (const auto &rule : qAsConst(d->mExRules)) {
                 allowed = allowed && !rule->recursAt(nextDT);
@@ -1245,8 +1239,8 @@ QDateTime Recurrence::getPreviousDateTime(const QDateTime &afterDateTime) const
         prevDT = dates.last();
 
         // Check if that date/time is excluded explicitly or by an exrule:
-        if (!std::binary_search(d->mExDates.constBegin(), d->mExDates.constEnd(), prevDT.date()) &&
-            !std::binary_search(d->mExDateTimes.constBegin(), d->mExDateTimes.constEnd(), prevDT)) {
+        if (!std::binary_search(d->mExDates.constBegin(), d->mExDates.constEnd(), prevDT.date())
+            && !std::binary_search(d->mExDateTimes.constBegin(), d->mExDateTimes.constEnd(), prevDT)) {
             bool allowed = true;
             for (const auto &rule : qAsConst(d->mExRules)) {
                 allowed = allowed && !rule->recursAt(prevDT);
@@ -1354,6 +1348,7 @@ void Recurrence::setRDateTimes(const QList<QDateTime> &rdates)
 
     d->mRDateTimes = rdates;
     sortAndRemoveDuplicates(d->mRDateTimes);
+    d->mRDateTimePeriods.clear();
     updated();
 }
 
@@ -1367,6 +1362,22 @@ void Recurrence::addRDateTime(const QDateTime &rdate)
     updated();
 }
 
+void Recurrence::addRDateTimePeriod(const Period &period)
+{
+    if (d->mRecurReadOnly) {
+        return;
+    }
+
+    setInsert(d->mRDateTimes, period.start());
+    d->mRDateTimePeriods.insert(period.start(), period);
+    updated();
+}
+
+Period Recurrence::rDateTimePeriod(const QDateTime &rdate) const
+{
+    return d->mRDateTimePeriods.value(rdate);
+}
+
 DateList Recurrence::rDates() const
 {
     return d->mRDates;
@@ -1460,35 +1471,35 @@ void Recurrence::dump() const
     int i;
     int count = d->mRRules.count();
     qCDebug(KCALCORE_LOG) << "  -)" << count << "RRULEs:";
-    for (i = 0;  i < count;  ++i) {
+    for (i = 0; i < count; ++i) {
         qCDebug(KCALCORE_LOG) << "    -) RecurrenceRule: ";
         d->mRRules[i]->dump();
     }
     count = d->mExRules.count();
     qCDebug(KCALCORE_LOG) << "  -)" << count << "EXRULEs:";
-    for (i = 0;  i < count;  ++i) {
+    for (i = 0; i < count; ++i) {
         qCDebug(KCALCORE_LOG) << "    -) ExceptionRule :";
         d->mExRules[i]->dump();
     }
 
     count = d->mRDates.count();
     qCDebug(KCALCORE_LOG) << "  -)" << count << "Recurrence Dates:";
-    for (i = 0;  i < count;  ++i) {
+    for (i = 0; i < count; ++i) {
         qCDebug(KCALCORE_LOG) << "    " << d->mRDates[i];
     }
     count = d->mRDateTimes.count();
     qCDebug(KCALCORE_LOG) << "  -)" << count << "Recurrence Date/Times:";
-    for (i = 0;  i < count;  ++i) {
+    for (i = 0; i < count; ++i) {
         qCDebug(KCALCORE_LOG) << "    " << d->mRDateTimes[i];
     }
     count = d->mExDates.count();
     qCDebug(KCALCORE_LOG) << "  -)" << count << "Exceptions Dates:";
-    for (i = 0;  i < count;  ++i) {
+    for (i = 0; i < count; ++i) {
         qCDebug(KCALCORE_LOG) << "    " << d->mExDates[i];
     }
     count = d->mExDateTimes.count();
     qCDebug(KCALCORE_LOG) << "  -)" << count << "Exception Date/Times:";
-    for (i = 0;  i < count;  ++i) {
+    for (i = 0; i < count; ++i) {
         qCDebug(KCALCORE_LOG) << "    " << d->mExDateTimes[i];
     }
 }
@@ -1504,12 +1515,14 @@ KCALENDARCORE_EXPORT QDataStream &KCalendarCore::operator<<(QDataStream &out, KC
     }
 
     serializeQDateTimeList(out, r->d->mRDateTimes);
+    out << r->d->mRDateTimePeriods.size();
+    for (auto it = r->d->mRDateTimePeriods.cbegin(); it != r->d->mRDateTimePeriods.cend(); ++it) {
+        out << it.key() << it.value();
+    }
     serializeQDateTimeList(out, r->d->mExDateTimes);
     out << r->d->mRDates;
     serializeQDateTimeAsKDateTime(out, r->d->mStartDateTime);
-    out << r->d->mCachedType
-        << r->d->mAllDay << r->d->mRecurReadOnly << r->d->mExDates
-        << r->d->mExRules.count() << r->d->mRRules.count();
+    out << r->d->mCachedType << r->d->mAllDay << r->d->mRecurReadOnly << r->d->mExDates << r->d->mExRules.count() << r->d->mRRules.count();
 
     for (RecurrenceRule *rule : qAsConst(r->d->mExRules)) {
         out << rule;
@@ -1528,15 +1541,25 @@ KCALENDARCORE_EXPORT QDataStream &KCalendarCore::operator>>(QDataStream &in, KCa
         return in;
     }
 
-    int rruleCount, exruleCount;
+    int rruleCount;
+    int exruleCount;
+    int size;
 
     deserializeQDateTimeList(in, r->d->mRDateTimes);
+    in >> size;
+    r->d->mRDateTimePeriods.clear();
+    r->d->mRDateTimePeriods.reserve(size);
+    for (int i = 0; i < size; ++i) {
+        QDateTime start;
+        Period period;
+        in >> start >> period;
+        r->d->mRDateTimes << start;
+        r->d->mRDateTimePeriods.insert(start, period);
+    }
     deserializeQDateTimeList(in, r->d->mExDateTimes);
     in >> r->d->mRDates;
     deserializeKDateTimeAsQDateTime(in, r->d->mStartDateTime);
-    in >> r->d->mCachedType
-       >> r->d->mAllDay >> r->d->mRecurReadOnly >> r->d->mExDates
-       >> exruleCount >> rruleCount;
+    in >> r->d->mCachedType >> r->d->mAllDay >> r->d->mRecurReadOnly >> r->d->mExDates >> exruleCount >> rruleCount;
 
     r->d->mExRules.clear();
     r->d->mRRules.clear();
